Require Export Program Classes.Equivalence Classes.Morphisms Setoid Relation_Definitions Classes.RelationClasses.

Set Implicit Arguments.

(** * Introduction: functional extensional equivalence *)

(** In Coq, [=] does not identify functions that have the same
signature but a different internal behaviour, for example two
different sorting algorithms.  It is possible to add the axiom of
proof irrevelance to Coq to identify such functions, or to reason
using an equivalence relation [=phi=] between functions. *)

(** * Equivalence definition, notation and properties *)

Definition funext_eq(A B:Type)(f g:A->B) := 
  forall a, f a = g a.
Notation "f =phi= g" := (funext_eq f g)(at level 80).

Program Instance funext_equiv (A B:Type): Equivalence (@funext_eq A B).
Solve All Obligations using (intros; auto with * || congruence).

Open Scope signature_scope.

Program Instance proper_compose_funext_eq {A B C:Type}: 
  Proper ((@funext_eq B C) ==>(@funext_eq A B)==>(@funext_eq A C)) (@compose A B C).
Next Obligation.
  intros f g H f' g' H' a.
  unfold compose. rewrite H, H'.
  trivial.
Qed.

Close Scope signature_scope.