Require Import Coq.Lists.List.
Import ListNotations.
Require Import Program.
Require Import ProgramCalculation.library.Notations.
Require Import ProgramCalculation.library.Tactics.
Require Import ProgramCalculation.theories.Algebra.Algebra.
Require Import ProgramCalculation.theories.Lists.Notations.
Require Import ProgramCalculation.theories.Lists.Induction.
Require Import ProgramCalculation.theories.Lists.Reduce.
Require Import ProgramCalculation.theories.Lists.Map.
Require Import ProgramCalculation.library.FunctionalExtensionalEq.
Generalizable All Variables.

(** * Rules about map *)
Section Map.

  (** ** empty rule *)
  Lemma empty_map `(f:A -> B) :
    (f ✶) ∘ (K []) =phi= K [] (B:=A).
  Proof. intro a; auto. Qed.

  (** ** one-point rule *)
  Lemma one_point_map `(f: A -> B) :
    f✶ ∘ [⋅] =phi= [⋅] ∘ f.
  Proof. intro a; auto. Qed.

  (** ** promotion rule *)
  Lemma map_promotion `(f: A -> B) :
    f✶ ∘ (++)/ =phi= (++) / ∘ (f✶)✶ .
  Begin.
    intro ll. 
    induction ll as [ | l | 
                      ll1 ll2 IHll1 IHll2 ] using join_ind.
    - trivial.
    - trivial.
    - check ( ( (f ✶) ∘ ((++) /) ) (ll1 ++ ll2) = 
              ( ((++) /) ∘ ((f ✶) ✶) ) (ll1 ++ ll2) ) is GOAL.
      LHS 
      = { by def compose; rewrite reduce_join }
      ( (f ✶) ( ((++)/ll1) ++ ((++)/ll2) ) ).
      = { rewrite map_app }
      ( ( (f ✶) ∘ ((++)/) ) ll1 ++  ( (f ✶) ∘ ((++)/) ) ll2 ).
      = { rewrite IHll1, IHll2 }
      ( ((++) / ((f ✶) ✶ ll1)) ++ ((++)/ ((f ✶) ✶ ll2) ) ).
      = { rewrite reduce_join }
      ( (++)/ ( ((f ✶) ✶ ll1) ++ ((f ✶) ✶ ll2) ) ).
      = { by def compose; rewrite map_app }
      ( (((++)/ ∘ ((f ✶) ✶)) (ll1 ++ ll2) ) ).
      □.
  Qed.

End Map.

(** * Rules about reduce *)
Section Reduce.
  
  Hypothesis A : Type.
  Hypothesis op : A -> A -> A.
  Hypothesis e : A.
  Infix "⊕" := op (at level 20).
  Notation "(⊕)" := (fun x y=>x⊕y).
  Notation "id⊕" := e.
  Context `{m:Monoid A op id⊕}.
  
  (** ** empty rule *)
  Lemma empty_reduce :
    (⊕) / ∘ K[] =phi= K id⊕ (B:=A).
  Proof. intro. auto. Qed.

  (** ** one-point rule *)
  Lemma one_point_reduce :
    (⊕) / ∘ [⋅] =phi= id.
  Begin.
    intro a.
    LHS
    = { by def compose; rewrite reduce_one_point }
    ( a ).
    □.
  Qed.
  
  (** ** promotion rule *)
  Lemma reduce_promotion : 
    (⊕)/ ∘ (++)/ =phi= (⊕)/ ∘ ((⊕)/)✶ .
  Begin.
    intro ll. 
    induction ll as [ | l | 
                      ll1 ll2 IHll1 IHll2 ] using join_ind.
    - trivial.
    - check ( ((⊕)/ ∘ (++)/) [l] = 
              ((⊕)/ ∘ ((⊕)/✶)) [l] ) is GOAL.
      LHS
      = { by def compose; rewrite reduce_one_point }
      ( (⊕)/ l ).
      RHS 
      = { by def compose; by def map }
      ( (⊕) / [ (⊕) / l ] ).
      = { rewrite reduce_one_point }
      ( (⊕)/ l ).
      □.
    - check( ((⊕) / ∘ (++) /) (ll1 ++ ll2) = 
             ((⊕) / ∘ ((⊕) /) ✶) (ll1 ++ ll2) ) is GOAL.
      LHS
      = { by def compose; repeat rewrite reduce_join }
      ( ((⊕) / ((++)/ ll1)) ⊕ ((⊕) / ((++)/ ll2)) ).
      = { by def compose; rewrite IHll1, IHll2 }
      ( (((⊕) / ∘ ((⊕) /) ✶) ll1) ⊕ (((⊕) / ∘ ((⊕) /) ✶) ll2) ). 
      = { by def compose; repeat rewrite reduce_join }
      ( (⊕) / ( ((⊕) / ✶ ll1) ++ ((⊕)/✶ ll2) ) ).
      = { by def compose; rewrite map_app }
      ( ((⊕) / ∘ ((⊕) /) ✶) (ll1 ++ ll2) ).
      □.
  Qed.
End Reduce.

(** * map and reduce promotion *)
Section Promotion.
  
  Hypothesis A : Type.
  Hypothesis op : A -> A -> A.
  Hypothesis e : A.
  Infix "⊕" := op (at level 20).
  Notation "(⊕)" := (fun x y=>x⊕y).
  Notation "id⊕" := e.
  Context `{m:Monoid A op id⊕}.

  Lemma map_and_reduce_promotion `(f:B -> A)`(g:C -> list B) : 
    (⊕)/ ∘ f✶ ∘ (++)/ ∘ g✶ =phi= (⊕)/ ∘ ((⊕)/ ∘ f✶ ∘ g)✶ .
  Begin.
    LHS 
    = { rewrite compose_assoc }
    (  (⊕)/ ∘ ( f✶ ∘ (++)/ ) ∘ g✶ ).
    = { rewrite map_promotion }
    ( (⊕) / ∘ (++)/ ∘ f✶✶ ∘ g✶ ).
    = { rewrite reduce_promotion } 
    ( (⊕)/ ∘ ((⊕)/✶) ∘ ((f✶)✶) ∘ (g✶) ).
    = { rewrite compose_assoc }
    ( ( (⊕)/ ∘ (⊕)/✶ ) ∘ ( ((f✶)✶) ∘ (g✶) ) ).
    = { rewrite map_distributivity }
    ( ( (⊕)/ ∘ ((⊕)/✶) ∘ (f✶ ∘ g)✶ ) ).
    = { rewrite compose_assoc }
    ( ( (⊕)/ ∘ ( (⊕)/✶ ∘ (f✶ ∘ g)✶ ) ) ).
    = { rewrite map_distributivity with (f:=(⊕)/)}
    ( (⊕)/ ∘ ( (⊕)/ ∘ (f✶ ∘ g))✶  ).
    □.
  Qed.
    
End Promotion.
